// Author : S.NICAISSE 
// VERSION 1.2
// This file is described the communicatives functions between master process and slaves process in a cluster. 
// Each slave process use the astronomical catalogue's software (for example : the pmm.exe for USNO_A2 catalogue).
// I use the MPI library

#include <stdio.h>
#include <mpi.h>  /* MPI library */
#include <string.h>
#include <stdlib.h> /* malloc */
#include "memory_catch.h" /* the dictionnary */
#include "work_catch.h" /* catch of the works */
#include "config_file.h"
#include <unistd.h>
#include <fcntl.h>
#include <sys/stat.h>
#include <signal.h>

#define MESSAGE 1
#define WAIT 1
#define NOWAIT 0
#define MASTER 0
#define SIZE_CONTROL 100
#define SIZE_RESULT 1024
#define NB_SLAVES 5
#define SIZE_ARRAY_WORK 65000


FILE *in;
int nbr_controls = 0, fd;

char *give_catalogue_name(char *control) {
  // give the catalogue's name of control.
  char *name, character = 'a';
  int index = 0;
  // printf("control=%s;\n", control);
  name = (char *)(malloc(SIZE_CONTROL*sizeof(char)));
  if (!name) {
    fprintf(stderr, "Not enough memory to name in dispatcher.\n");
    exit(0);
  }
  character = *(control + index);
  while (character != ' ') {
    *(name + index++) = character;
    character = *(control + index);
  }
  *(name + index) = '\0';
  return(name);
}


int give_control(char *control, char *name_pipe) {
  // give a control and set in control buffer. The groupe_controls symboles the groupes to extract. Index symbole the ptre where the new
  // groupe is located in the groupe_controls buffer.
  // If the extract is good then return 1 else 0.
  // printf("give_control\n");
  fread(control, SIZE_CONTROL, sizeof(char), in);
  nbr_controls++;
  /* if (nbr_controls % 1000 == 0) {
    fclose(in);
    unlink(name_pipe);
    fprintf(stderr, "destruction du pipe \n");
    if (mkfifo(name_pipe, 0600)==-1) {   /* creation du tube 
    perror("dispatcher : Build a name pipe."); 
      exit(1); 
    }
    fprintf(stderr, "le make fonctionne \n");  
    if ((fd=open(name_pipe, O_RDONLY))==-1) { /* ouverture du tube en lecture 
      perror("dispatcher : Open the name pipe"); 
      exit(1); 
    } 
    fprintf(stderr, "pipe = %s\n", name_pipe);
    fflush(stderr);
    
    if ((in=fdopen(fd, "r"))==NULL) { /* ouverture d'un flux associe a fd 
      perror("dispatcher : open the stream of fd"); 
      exit(1); 
    } 
    nbr_controls = 0;
    fprintf(stderr, "envoi signal\n");
    kill(0, SIGHUP);
  } */
  /* fprintf(stderr, "control = %s\n", control);
     fflush(stderr); */
  return(strncmp(control, "finish", 6));
}

void cut_string(char *string_array[100], char *string) {
  // Cut a string where there is a space character. It place the new string in the array. it begin at index = 1 because it doesn't 
  // know the executable file.
  char *string_result, character = 'a';
  int index = 0, index_array = 1, index_result;
  // printf("cut_string.\n");
  while(character != '\0') {
    character = *(string + index++);
    string_result = (char *)(malloc(SIZE_CONTROL*sizeof(char)));
    if (!string_result) {
      fprintf(stderr, "Not enough memory string_result of dispatcher.\n");
      exit(1);
    }
    string_array[index_array++] = string_result;
    index_result = 0;
    while(character != '\0' && character != ' ') {
      *(string_result + index_result++) = character;
      character =  *(string + index++);
    }
    *(string_result + index_result) = '\0';
  }
  string_array[index_array] = NULL;
  // printf("index_array = %d", index_array);
  index = 0;
}


int main(int argc, char *argv[]) { 

  
  double startwtime, endwtime; 
  int namelen, numprocs, myid, err, flag, slave_terminate = 0, end = 0, i = 1, slave_state[NB_SLAVES] ;
  long id_job = 1, print_job = 1;
  int id_result, give_result, result = 0;
  char processor_name[MPI_MAX_PROCESSOR_NAME];
  char *buff, *a_control, *locate_slaves[NB_SLAVES], *name_pipe;

  MPI_Request request;
  MPI_Status status;
  // printf("argc\n");
  MPI_Init(&argc, &argv); 
  // printf("init\n");
  MPI_Comm_size(MPI_COMM_WORLD, &numprocs); 
  MPI_Comm_rank(MPI_COMM_WORLD, &myid); 
  MPI_Get_processor_name(processor_name,&namelen);
  MPI_Barrier(MPI_COMM_WORLD);
  // printf("bonjour\n");
  // fflush(stdout);
  init_config_file();
  // show_informations();
  // printf ("debut fin %d\n", fin);
  // printf("esclave %d nom %s\n", myid, processor_name);
  if (myid==0) {
    // Master
    // printf("je construit les controles.\n");
    // system("sleep 5");
    // printf("j'envoie les controls.\n");
    // initialisation of dictionnary.
    name_pipe = (char *)(malloc(100*sizeof(char)));
    init_memory_catch();
    // init_config_file();
    init_work_catch();
    // init the states of slave. If state  > 0 shows the id of commande.
    // If state = 0 then slave wait a job.
    // If state = -1 then slave is dead.
    /* for(i=0 ; i < argc ; i++) {
       printf("argc = %d argv[%d]=%s\n", argc, i, argv[i]);
       } */
    // fprintf(stderr, "dispatcher : nom du pipe : %s\n", argv[1]);
    // fflush(stderr);
    strcpy(name_pipe, argv[1]);
    if (mkfifo(name_pipe, 0600)==-1) {   /* creation du tube */ 
      perror("dispatcher : Build a name pipe."); 
      exit(1); 
    }  
    if ((fd=open(name_pipe, O_RDONLY))==-1) { /* ouverture du tube en lecture */ 
      perror("dispatcher : Open the name pipe"); 
      exit(1); 
    } 
    // fprintf(stderr, "pipe = %s\n", argv[1]);
    // fflush(stderr);
    
    if ((in=fdopen(fd, "r"))==NULL) { /* ouverture d'un flux associe a fd */ 
      perror("dispatcher : open the stream of fd"); 
      exit(1); 
    } 
    for(i = 0 ; i < NB_SLAVES ; i++) {
      // The master answer to slave. Where are you ? A slave tell : I'm in cocat2.u-strasbg.fr.
      // printf("fin boucle = %d\n", fin);
      slave_state[i] = 0;
      buff = (char *)(malloc(MPI_MAX_PROCESSOR_NAME*sizeof(char)));
      if (!buff) {
	fprintf(stderr, "Not enough memory to a_control buffer of dispatcher.\n");
	exit(1);
      }
      MPI_Isend((void *)("reponse"), 10*sizeof(char), MPI_CHAR, i+1, MESSAGE, MPI_COMM_WORLD, &request); 
      MPI_Wait(&request, &status);
      MPI_Irecv(buff, MPI_MAX_PROCESSOR_NAME*sizeof(char), MPI_CHAR, i+1, MESSAGE, MPI_COMM_WORLD, &request); 
      MPI_Wait(&request, &status);
      locate_slaves[i] = buff;
    }
    do {
      // fprintf(stderr, "in = %d\n", fileno(in));
      // fflush(stderr);
      // printf("autre fin = %d\n", fin);
      for(i = 1 ; i < numprocs ; i++) {
	if (slave_state[i - 1] == 0) {
	  // The slave wants to work.
	  a_control = (char *)(malloc(SIZE_CONTROL*sizeof(char)));
	  if (!a_control) {
	    fprintf(stderr, "Not enough memory to a_control buffer of dispatcher.\n");
	    exit(1);
	  }
	  
	  if (end) {

	    // There aren't any jobs.
	    // printf("terminate ; end = %d\n", fin);
	    strcpy(a_control, "finish");
	    slave_state[i - 1] = -1;
	    slave_terminate++;
	  }
	  else {
	    result = 0;
	    int nb_works = number_works();
	    if (nb_works > 0) {
	      // There are the works in the work catch.
	      Work *a_work;
	      int index_work = 0;
	      int nbr_works = nb_works;
	      char *catalogue_name;
	      while (nbr_works > 0 && result == 0) {
		// printf("Cherche du travail.\n");
		a_work = give_work(&index_work);
		catalogue_name = give_catalogue_name(a_work -> control);
		result = can_work_catalogue(catalogue_name, locate_slaves[i-1]);
		free(catalogue_name);
		nbr_works--;
	      }
	      if (result) {
		free(a_control);
		a_control = a_work -> control;
		slave_state[i - 1] = a_work -> nbr_control;
		delete_work(slave_state[i - 1]);
	      }
	    }
	    else {
	      while(!result && !end) {
		// printf("index = %d.\n", index);
		give_result = give_control(a_control, name_pipe);
		if (!give_result) { 
		  // There aren't any jobs.
		  // This condition isn't including in the "end" condition because the "give_control" function read the file. But,
		  // the "end" shows the cursor is the end of file. If this function is using with this state then the software bugs.
		  // printf("finish : argv[1] = %s\n", argv[index_argument] + index);
		  strcpy(a_control, "finish");
		  slave_state[i - 1] = -1;
		  slave_terminate++;
		  end = 1;
		}
		else {
		  if (!give_result) {
		    // I take an another argument.
		    give_control(a_control, name_pipe);
		  }
		  // Can this slave work with this catalogue ?
		  char *catalogue_name;
		  // printf("a_control = %s.\n", a_control);
		  catalogue_name = give_catalogue_name(a_control);
		  // printf("catalogue_name = %s.\n", catalogue_name);
		  // show_informations();
		  result = can_work_catalogue(catalogue_name, locate_slaves[i-1]);
		  // printf("result = %d avec le catalogue %s pour l'esclave %d situ au %s\n", result, catalogue_name, i, locate_slaves[i-1]);
		  if (!result) {
		    // Set control in work_catch.
		    add_new_buffer(id_job);
		    add_work(a_control, id_job++);
		    a_control = (char *)(malloc(SIZE_CONTROL*sizeof(char)));
		    if (!a_control) {
		      fprintf(stderr, "Not enough memory to a_control buffer of dispatcher.\n");
		      exit(1);
		    }
		  }
		  free(catalogue_name);
		}
	      }
	      if (result) {
		// The slave has a work.
		add_new_buffer(id_job);
		// printf("id_job = %d.\n", id_job);
		slave_state[i - 1] = id_job++;
	      }
	    }
	  }
	  // Send the control.
	  // printf("a_control = %s\n", a_control);
	  if (end || result) {
	    MPI_Isend(a_control, SIZE_CONTROL*sizeof(char), MPI_CHAR, i, MESSAGE, MPI_COMM_WORLD, &request); 
	    MPI_Wait(&request, &status);
	  }
	  free(a_control);
	}
      }
      for (i = 1 ; i < numprocs ; i++) {
	if (slave_state[i - 1] > 0) {
	  MPI_Isend((void *)("reponse"), 10*sizeof(char), MPI_CHAR, i, MESSAGE, MPI_COMM_WORLD, &request); 
	  MPI_Test(&request, &flag, &status);
	  int nb_job = slave_state[i - 1];
	  if (flag) {
	    id_result = 0;
	    do {
	      buff = (char *)(malloc(SIZE_RESULT*sizeof(char)));
	      if (!buff) {
		fprintf(stderr, "Not enough memory to buff buffer of dispatcher (master).\n");
		exit(1);
	      }
	      MPI_Irecv(buff, SIZE_RESULT*sizeof(char), MPI_CHAR, i, MESSAGE, MPI_COMM_WORLD, &request); 
	      MPI_Wait(&request, &status);
	      if (strcmp(buff, "finish"))
		put_string(nb_job, id_result++, buff);
	      // printf("nb_job = %d.\n", nb_job);
	      // printf("%s", buff);
	    } while (strcmp(buff, "finish"));
	    slave_state[i - 1] = 0;
	  }
	  else {
	    printf("le message reponse ne passe pas.\n");
	  }
	}
      }    
      // Print the results.
      int stop = 0;
      for (i = print_job ; i < id_job && !stop  ; i++) {
	id_result = 0;
	// printf("reponse %d.\n", i);
	buff = get_string(i, id_result++);
	stop = (buff == NULL);
	while (buff != NULL) {
	  printf("%s", buff);
	  buff = get_string(i, id_result++);
	}
	if (!stop) 
	  delete_string(i);
      }
      print_job = i;
    } while(slave_terminate < NB_SLAVES);
    fclose(in);
    unlink(argv[1]);  
  } 
  else {
    // The slaves.
    char buffer[10], autre_buff[1024], *array_control[100], *chaine;
    int i=0, tube[2]; 
    char ch[1024], *catalogue_name;
    Information *an_information;
    MPI_Irecv(buffer, 10*sizeof(char), MPI_CHAR, MASTER, MESSAGE, MPI_COMM_WORLD, &request); 
    MPI_Wait(&request, &status); 
    // send the node where the slave is located.
    MPI_Isend(processor_name, MPI_MAX_PROCESSOR_NAME*sizeof(char), MPI_CHAR, MASTER, MESSAGE, MPI_COMM_WORLD, &request); 
    MPI_Wait(&request, &status);
    buff = (char *)(malloc(SIZE_CONTROL*sizeof(char)));
    do {
      free(buff);
      buff = (char *)(malloc(SIZE_CONTROL*sizeof(char)));
      if (!buff) {
	fprintf(stderr, "Not enough memory to buff buffer of dispatcher (slave).\n");
	exit(1);
      }
      MPI_Irecv(buff, SIZE_CONTROL*sizeof(char), MPI_CHAR, MASTER, MESSAGE, MPI_COMM_WORLD, &request); 
      MPI_Wait(&request, &status);
      if (strcmp(buff, "finish")) {
	// printf("on lance pmm.exe.\n");
	// printf("chaine de controle %s.\n", buff + 1); 
	if (pipe(tube) == -1) { 
	  perror("Ouverture pipe impossible"); 
	  exit(1); 
	} 
	
	
	i=fork(); 
	
	if (i==0) {
	  /* process fils */ 
	  close(1); /* fermeture sortie std */ 
          catalogue_name = give_catalogue_name(buff);
	  // printf("catalogue_name fils = %s.\n", catalogue_name);
	  an_information = give_information(catalogue_name);
	  // printf("an_information = %d.\n", an_information);
	  chaine = (char *)(malloc(10*sizeof(char)));
	  // printf("executable_file = %s.\n", give_executable_file(an_information));
	  strcpy(chaine, give_executable_file(an_information));
	  array_control[0] = chaine;
	  // printf("controle execute =%s\n", buff + strlen(catalogue_name) + 1);
	  cut_string(array_control, buff + strlen(catalogue_name) + 1);
	  dup2(tube[1], 1);  /* redirection sortie std sur descripteur en ecriture du tube */ 
	  close(tube[0]);  /* fermeture du descripteur inutilise */ 
	  // close (tube[1]); /* inutile de garder 2 descripteurs ouverts pour le meme fichier 
	  // tout acces sur la sortie std va deja  dans le tube */ 
	 
	  execv(give_path_executable_file(an_information), array_control); 
	  // execv("/bin/ls", array_control);
	  perror("Lancement mpirun"); 
	} 

	else { 
	  int n;
	  /* process pere */ 
	  close (tube[1]); 
	  // if ((in=fdopen(tube[0], "r"))==NULL) { /* ouverture d'un flux associe a tube[0] plus pratique pour la lecture formatte*/ 
	  /* perror("ouverture tube"); 
	     exit(1); 
	     } */ 
	  MPI_Irecv(buffer, 10*sizeof(char), MPI_CHAR, MASTER, MESSAGE, MPI_COMM_WORLD, &request); 
	  MPI_Wait(&request, &status);
	  while((n = read(tube[0], ch, 1023))) {
	    ch[n] = 0;
	    MPI_Isend(ch, 1024*sizeof(char), MPI_CHAR, MASTER, MESSAGE, MPI_COMM_WORLD, &request); 
	    MPI_Wait(&request, &status);
	    // fflush(0);
	  }
	  MPI_Isend((void *)("finish"), 1024*sizeof(char), MPI_CHAR, MASTER, MESSAGE, MPI_COMM_WORLD, &request); 
	  MPI_Wait(&request, &status);
	}
	wait(0); 
	close(tube[0]);
      }
 
    } while (strcmp(buff, "finish"));
  } 
  // system("sleep 100");
  MPI_Finalize();

  return 0;


}
